/*
    ScriptExec - binary bundled into Platypus-created applications
    Copyright (C) 2006 Sveinbjorn Thordarson <sveinbjornt@simnet.is>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*

	I would like to thank Gianni Ceccarelli for his contribution of modified code
	which enabled piping from a privileged task into a Text Window

*/

#import "ScriptExecController.h"
#include <Security/Authorization.h>
#include <sys/wait.h>
#include <stdio.h>
#include <fcntl.h>

@implementation ScriptExecController

- (id)init
{
	if (self = [super init]) 
	{
		//create arrays
		arguments = [[NSMutableArray alloc] initWithCapacity: kMaxFileArguments+2];
		fileArgs = [[NSMutableArray alloc] initWithCapacity: kMaxFileArguments];
		isTaskDone = NO;
		isTaskStarted = NO;
    }
    return self;
}

-(void)dealloc
{
	if (arguments != NULL) { [arguments release]; }
	if (fileArgs != NULL)  { [fileArgs release];  }
	
	[super dealloc];
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	//load settings from app bundle
	[self loadSettings];
	
	//put application name into the relevant menu items
	[quitMenuItem setTitle: [NSString stringWithFormat: @"Quit %@", appName]];
	[aboutMenuItem setTitle: [NSString stringWithFormat: @"About %@", appName]];
	[hideMenuItem setTitle: [NSString stringWithFormat: @"Hide %@", appName]];
	
	//create progress window if so specified
	if (outputType == kProgressBarOutput)
	{
		//prepare progress bar, then center and display window
		[progressBar setUsesThreadedAnimation: YES];
		[progressBar startAnimation: self];
		[progressWindow setTitle: appName];
		[progressWindow center];
		[progressWindow makeKeyAndOrderFront: NULL];
	}
	else if (outputType == kTextOutput)
	{   
		//prepare text window controls, then center and display the window
		[textOutputTextField setFont:[NSFont userFixedPitchFontOfSize: 10.0]];
		[textOutputProgressIndicator setUsesThreadedAnimation: YES];
		[textOutputProgressIndicator startAnimation: self];
		[textOutputWindow setTitle: appName];
		[textOutputWindow center];
		[textOutputWindow makeKeyAndOrderFront: NULL];
	}
	
	// If the script is encrypted, we need to decrypt it into a 'script' file in order to run it
	if (encryptAndChecksum)
	{
		NSString *decryptedScript = [SillyStringEncrypt sillyDecryptString: [NSString stringWithContentsOfFile: hiddenScriptPath]];
		[decryptedScript writeToFile: kTempScriptFile atomically: YES];
	}

	//create argument list
	if (encryptAndChecksum)
		[arguments addObject: kTempScriptFile];
	else
		[arguments addObject: scriptPath];
	
	[arguments addObject: [[NSBundle mainBundle] bundlePath]]; //$1 is path of application bundle
	if ([fileArgs count] > 0)//if there are any dropped files, we add them as arguments after $1
		[arguments addObjectsFromArray: fileArgs];

	//start new thread for executing script
	if (execStyle == kPrivilegedExecution) //Authentication mode
		[self executeScriptWithPrivileges];
	else //plain old regular
		[self executeScript];
}

#pragma mark -

//launch regular user task with NSTask
- (void)executeScript
{	
	isTaskStarted = YES;

	//initalize task
	task = [[NSTask alloc] init];
	
	if (outputType == kTextOutput)// we monitor output if TextWindow output is on
	{
		outputPipe = [NSPipe pipe];
		[task setStandardOutput: outputPipe];
		[task setStandardError: outputPipe];
		readHandle = [outputPipe fileHandleForReading];
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(getData:) name:NSFileHandleReadCompletionNotification object:readHandle];
		[readHandle readInBackgroundAndNotify];
	}
	
	//apply settings for task
	[task setLaunchPath: interpreter];
	[task setArguments: arguments];
	
	//set it off
	[task launch];
	
	//set off timer that checks task status, i.e. when it's done 
	[NSTimer scheduledTimerWithTimeInterval: 0.25 target: self selector:@selector(checkTaskStatus) userInfo: nil repeats: YES];
}

// check if task is running
- (void)checkTaskStatus
{
	if (![task isRunning])//if it's no longer running, we do clean up
		[self taskFinished];
}

#pragma mark -


//launch task with privileges using Authentication Manager
- (void)executeScriptWithPrivileges
{
	OSErr					err = noErr;
    short					i;
    char					*args[kMaxFileArguments+2];
	char					interpreterStr[kMaxPathLength];
	FILE					*outputFile;
	
	isTaskStarted = YES;
	
	//interpreter
	[interpreter getCString: (char *)&interpreterStr maxLength: kMaxPathLength];
	
	//create arguments array
	for (i = 0; i < [arguments count]; i++)
	{
		args[i] = malloc(kMaxPathLength);
		[[arguments objectAtIndex:i] getCString: (char *)args[i] maxLength: kMaxPathLength];
		if (i == [arguments count]-1) // if last argument, we terminate array of args with a null
			args[i+1] = NULL;
	}
    
    // Use Apple's Authentication Manager APIs to get an Authorization Reference
	// This is not the Apple-recommended way of doing this -- but with setuid under attack etc, it'll just have to do
	// The question remains whether we should use something other than an empty environment
	// A lot of people have reported problems with scripts because of unset environmental variables
    err = AuthorizationCreate(NULL, kAuthorizationEmptyEnvironment, kAuthorizationFlagDefaults, &authorizationRef);
    if (err != errAuthorizationSuccess)
	{
		NSLog(@"Authorization for script execution failed - Error %d", err);
        [[NSApplication sharedApplication] terminate: self];
	}
	
	//use Authorization Reference to execute script with privileges
    if (!(err = AuthorizationExecuteWithPrivileges(authorizationRef,(char *)&interpreterStr, kAuthorizationFlagDefaults, args, &outputFile)) != noErr)
	{
		//get NSFileHandle for the task output
		readHandle = [[NSFileHandle alloc] initWithFileDescriptor: fileno(outputFile)];
		childPid = fcntl(fileno(outputFile), F_GETOWN, 0);//get pid
		
		// read the filehandle in the background and regularly print output into text window
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(getData:) name:NSFileHandleReadCompletionNotification object:readHandle];
		[readHandle readInBackgroundAndNotify];
		[NSTimer scheduledTimerWithTimeInterval: 0.25 target: self selector:@selector(checkPrivilegedTaskStatus) userInfo: nil repeats: YES];
	}
	else
	{
		NSLog(@"Error %d occurred when attempting to run AuthorizationExecuteWithPrivileges. Terminating...", err);
		[[NSApplication sharedApplication] terminate: self];
	}
}

// check if privileged task is running
- (void)checkPrivilegedTaskStatus
{
    int ret,pid;
    pid = waitpid(childPid, &ret, WNOHANG);
    if (pid != 0)
        [self taskFinished];
}

#pragma mark -

// OK, called when task is finished.  Some cleaning up to do, controls need to be adjusted, etc.
- (void)taskFinished
{
		//if we're using the hidden encrypted script, we must remove the temporary clear-text one in /tmp/
		if (encryptAndChecksum)
		{
			if ([[NSFileManager defaultManager] fileExistsAtPath: kTempScriptFile])
				[[NSFileManager defaultManager] removeFileAtPath: kTempScriptFile handler: nil];
		}
		
		
		if (!remainRunning)
		{	// we quit if the app isn't explicity set to continue running
			[[NSApplication sharedApplication] terminate: self];
		}
		else if (outputType == kTextOutput)
		{
			//update controls for text output window
			[textOutputCancelButton setTitle: @"Quit"];
			[textOutputCancelButton setKeyEquivalent:@"\r"];
			[textOutputProgressIndicator stopAnimation: self];
		}
		else if (outputType == kProgressBarOutput)
		{
			//update controls for progress bar output
			[messageTextField setStringValue: @"Task completed"];
			[progressBar stopAnimation: self];
			[cancelButton setTitle: @"Quit"];
			[cancelButton setKeyEquivalent:@"\r"];
		}
		
		isTaskDone = YES;
}


// read from the file handle and append it to the text window
- (void) getData: (NSNotification *)aNotification
{
	NSData *data;
	id notificationObject;

	//get the data
	notificationObject = [aNotification object];
	data = [[aNotification userInfo] objectForKey:NSFileHandleNotificationDataItem];
	
	//make sure there's actual data
	if ([data length]) 
	{
		//append the output
		outputStr = [[[NSString alloc] initWithData: data encoding: NSASCIIStringEncoding] autorelease];
		[textOutputTextField setString: [[textOutputTextField string] stringByAppendingString: outputStr] ];
		//scroll to it
		[textOutputTextField scrollRangeToVisible:NSMakeRange([[textOutputTextField string] length], 0)];
		// we schedule the file handle to go and read more data in the background again.
		[notificationObject readInBackgroundAndNotify];
	}
}



#pragma mark -

// respond to AEOpenDoc -- so MUCH more convenient than working with Apple Event Descriptors
- (BOOL)application:(NSApplication *)theApplication openFile:(NSString *)filename
{
	if (!isTaskStarted)
	{
		[fileArgs addObject: filename];
		return TRUE;
	}
	else //once script task has started, we refuse all opened files
		return FALSE;
	// This is something worth working on -- the launching of the script again when new
	// files are dropped
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	//terminate task
	if (task != NULL)
	{
		if ([task isRunning])
			[task terminate];
		[task release];
	}
	
	// just one more time, make sure we don't leave the clear-text script in the /tmp/ directory
	if ([[NSFileManager defaultManager] fileExistsAtPath: kTempScriptFile])
				[[NSFileManager defaultManager] removeFileAtPath: kTempScriptFile handler: nil];
	
	return(YES);
}

#pragma mark -

//load configuration files from application bundle
- (void)loadSettings
{
	NSBundle		*appBundle = [NSBundle mainBundle];
	NSFileManager   *fmgr = [NSFileManager defaultManager];
	NSDictionary	*appSettingsPlist;
	
	//make sure all the config files are present -- if not, we quit
	if (![fmgr fileExistsAtPath: [appBundle pathForResource:@"AppSettings.plist" ofType:nil]] || ![fmgr fileExistsAtPath: [appBundle pathForResource:@"script" ofType:nil]] || ![fmgr fileExistsAtPath: [appBundle pathForResource:@".script" ofType:nil]])
		[self fatalAlert: @"Corrupted app bundle" subText: @"Vital configuration and script files were missing from the application bundle."];
	
	//get app name
	appName = [[appBundle executablePath] lastPathComponent];
	
	//get dictionary with app settings
	appSettingsPlist = [NSDictionary dictionaryWithContentsOfFile: [appBundle pathForResource:@"AppSettings.plist" ofType:nil]];
	
	//determine output type
	NSString *outputTypeStr = [appSettingsPlist objectForKey:@"OutputType"];
	if ([outputTypeStr isEqualToString: @"ProgressBar"])
		outputType = kProgressBarOutput;
	else if ([outputTypeStr isEqualToString: @"TextWindow"])
		outputType = kTextOutput;
	else
		outputType = kNoOutput;
		
	//determine execution style
	execStyle = [[appSettingsPlist objectForKey:@"RequiresAdminPrivileges"] boolValue];

	//remain running?
	remainRunning = [[appSettingsPlist objectForKey:@"RemainRunningAfterCompletion"] boolValue];
	
	//is script encrypted and checksummed?
	encryptAndChecksum = [[appSettingsPlist objectForKey: @"EncryptAndChecksum"] boolValue];
	
	//get interpreter
	interpreter = [appSettingsPlist objectForKey:@"ScriptInterpreter"];
		
	//get path to script
	scriptPath = [appBundle pathForResource:@"script" ofType:nil];
	hiddenScriptPath = [appBundle pathForResource:@".script" ofType:nil];
	
	//if it requires checksum and is encryped, we verify checksum
	if (encryptAndChecksum)
	{
		// if it doesn't match the checksum, we exit program
		if (![STChecksum doesFileMatchChecksumString: hiddenScriptPath checksumString: [appSettingsPlist objectForKey: @"ScriptChecksum"]])
			[self fatalAlert: @"Invalid Checksum" subText: @"The script checksum does not match, which indicates that the script for this application is either corrupt or has been tampered with"]; 
		
		// make sure we can write to /tmp/ -- you never know, eh?
		if (![fmgr isWritableFileAtPath: @"/tmp/"])
		{
			[self fatalAlert: @"Unable to write temporary files" subText: @"Could not write to the /tmp/ directory.  Make sure this directory exists and that you have write privileges."]; 
			[[NSApplication sharedApplication] terminate: self];
		}
		
		// if for some reason, a temporary clear-text script exists in /tmp/, we remove it
		if ([fmgr fileExistsAtPath: kTempScriptFile])
				[fmgr removeFileAtPath: kTempScriptFile handler: nil];
	}
}



// Respond to Cancel by exiting application
- (IBAction)cancel:(id)sender
{
	[[NSApplication sharedApplication] terminate: self];
}

#pragma mark -

- (void)fatalAlert: (NSString *)message subText: (NSString *)subtext
{
	NSAlert *alert = [[NSAlert alloc] init];
	[alert addButtonWithTitle:@"OK"];
	[alert setMessageText: message];
	[alert setInformativeText: subtext];
	[alert setAlertStyle: NSCriticalAlertStyle];
	
	if ([alert runModal] == NSAlertFirstButtonReturn) 
	{
		[alert release];
		[[NSApplication sharedApplication] terminate: self];
	} 
}


@end
